@@ -7,6 +7,8 @@ use warnings;
 use Data::Dumper;
 $Data::Dumper::Indent = 1;
 
+our %env;
+
 sub _mangle_ident {
     my $s = shift;
     $s =~ s/ ([^a-zA-Z0-9_:]) / '_'.ord($1).'_' /xge;
@@ -14,11 +16,13 @@ sub _mangle_ident {
 }
 
 sub _mangle_var {
-    my $s = shift;
+    my $s = $_[0];
     #warn "mangle: $s";
     
     # perl6 => perl5 variables
     return '%::ENV'    if $s eq '%*ENV';  
+    return '$^O'       if $s eq '$*OS';  
+    
     # special variables
     return '$::_EXCL_' if $s eq '$!';
 
@@ -26,6 +30,41 @@ sub _mangle_var {
     return $s;
 }
 
+sub _var_get {
+    my $n = $_[0];
+    
+    if ( ! exists $n->{scalar} ) {
+        if ( exists $n->{bare_block} ) {
+            # TODO - check if it is a comma-delimited list
+            return ' sub ' . _emit( $n );
+        }
+        return _emit( $n );
+    }
+
+    my $s = $n->{scalar};
+
+    return $env{$s}{get}
+        if exists $env{$s} &&
+           exists $env{$s}{get};
+    
+    # default
+    return "\$self->{'" . substr($s,2) . "'}"
+        if substr($s,1,1) eq '.';
+    return _mangle_var( $s );
+}
+
+sub _var_set {
+    my $s = $_[0];
+    
+    #warn "emit: set $s - ", Dumper %env;
+    
+    return $env{$s}{set}
+        if exists $env{$s}{set};
+    
+    # default
+    return sub { _mangle_var( $s ) . " = " . $_[0] };
+}
+
 sub _not_implemented {
     my ( $n, $what ) = @_;
     return "die q(not implemented $what: " . Dumper( $n ) . ")";
@@ -34,20 +73,40 @@ sub _not_implemented {
 sub emit {
     
     # <audreyt> %Namespace:: = ();  # clear stash
+    local %env;
     
     my ($grammar, $ast) = @_;
     # runtime parameters: $grammar, $string, $state, $arg_list
     # rule parameters: see Runtime::Rule.pm
-    return _emit( $ast, '' );
+    #warn "Perl5::emit: ", Dumper( $ast ); 
+    return _emit( $ast );
         #"do{\n" .
         #_emit( $ast, '    ' ) . "\n" .
         #"}";
 }
 
+sub _emit_code {
+    my $code = $_[0];
+    if (substr($code, 1,1) eq '?') {
+        my $caller_level = 0;
+        while ($code =~ s/^&\?CALLER::/&?/) {
+            ++$caller_level;
+        }
+        my $name = substr($code, 2);
+        # special!
+        if ($name eq 'ROUTINE') {
+            return "Pugs::Runtime::Perl6::Routine->new(Devel::Caller::caller_cv($caller_level))";
+        }
+        die 'unhandled magic variable';
+    }
+
+    return $code;
+}
+
 sub _emit {
     my $n = $_[0];
     #die "_emit: ", Dumper( $n ); 
-    #warn "fixity: $n->{fixity}\n" if exists $n->{fixity};
+    #warn "_emit: ", Dumper( $n ); 
     
     # 'undef' example: parameter list, in a sub call without parameters
     return ''
@@ -55,11 +114,22 @@ sub _emit {
     
     die "unknown node: ", Dumper( $n )
         unless ref( $n ) eq 'HASH';
-        
+
+
+    return join ( ";\n", 
+            map { _emit( $_ ) } @{$n->{statements}} 
+        ) ||
+        " # empty block\n"
+        if exists $n->{statements};
+    
+
     return _mangle_ident( $n->{bareword} )
         if exists $n->{bareword};
+
+    return _mangle_ident( $n->{dot_bareword} )
+        if exists $n->{dot_bareword};
         
-    return $n->{code} 
+    return _emit_code($n->{code})
         if exists $n->{code};
         
     return $n->{int} 
@@ -68,7 +138,10 @@ sub _emit {
     return $n->{num} 
         if exists $n->{num};
         
-    return _mangle_var( $n->{scalar} )
+    return '{' . _emit( $n->{pair}{key} ) . '=>' . _emit( $n->{pair}{value} ) . '}'
+        if exists $n->{pair};
+        
+    return _var_get( $n )
         if exists $n->{scalar};
         
     return _mangle_var( $n->{array} )
@@ -100,10 +173,12 @@ sub _emit {
             if $n->{fixity} eq 'circumfix';
         return postcircumfix( $n )
             if $n->{fixity} eq 'postcircumfix';
+        return ternary( $n )
+            if $n->{fixity} eq 'ternary';
     }
     
     return statement( $n )
-        if ref $n->{op1} && exists $n->{op1}{stmt};
+        if exists $n->{statement};
 
     return default( $n );
 }
@@ -124,7 +199,7 @@ sub assoc_list {
 
 sub _emit_parameter_binding {
     my $n = $_[0];
-    
+
     # no parameters
     return ''
         if  ! defined $n ||
@@ -132,7 +207,7 @@ sub _emit_parameter_binding {
     
     #warn "parameter list: ",Dumper $n;
 
-    #    'scalar' => '$desc',    $v
+    #    'name' => '$desc',      $v
     #    'optional' => 1,        $v?
     #    'named_only' => 1,      :$v
     #    'type' => {             Str $v
@@ -141,16 +216,38 @@ sub _emit_parameter_binding {
     #    'splat' => 1,           *$v
     #    'attribute' => \@attr   $v is rw
 
-    my $param = join( ',' , 
-        map { _emit( $_ ) } @$n
-    );
-    return "my ($param) = \@_;\n";
+    # XXX: This is ugly, but nessary to support lexical subs passed it.
+    # This will go away along with Data::Bind integration, hopefully.
+    my $bind_code = '';
+    #warn "binding ", Dumper $n;
+    for (@$n) {
+        #warn "binding ", Dumper $_;
+        next unless ref $_;
+        if ( exists $_->{type} && $_->{type} eq 'Code' ) {
+            my $name = $_->{name};
+            $name =~ s/^&//;
+            $bind_code .= "   local *$name = shift;\n";
+        }
+        else {
+            $bind_code .= "   my ".$_->{name}." = shift;\n";
+        }
+    }
+    return $bind_code;
 }
 
 sub default {
     my $n = $_[0];
     #warn "emit: ", Dumper( $n );
     
+    if ( exists $n->{die} ) {
+        return  "do { die '" . $n->{die} . "' }";
+    }
+
+    if ( exists $n->{pointy_block} ) {
+        # XXX: no signature yet
+        return  "sub {\n" . _emit( $n->{pointy_block} ) . "\n }\n";
+    }
+
     if ( exists $n->{bare_block} ) {
         if ( exists $n->{trait} ) {
             # BEGIN/END
@@ -167,6 +264,7 @@ sub default {
             # Moose: package xxx; use Moose;
             # class Point;
             #warn "class: ",Dumper $n;
+            local %env;
             my $id;
             $id = exists $n->{param}{cpan_bareword} 
                   ? _mangle_ident( $n->{param}{cpan_bareword} )
@@ -179,6 +277,24 @@ sub default {
                 ";use Exporter 'import';our \@EXPORT";
         }
 
+        if ( $n->{sub}{bareword} eq 'is' ) {
+            # is Point;
+            #warn "inheritance: ",Dumper $n;
+            my $id;
+            $id = exists $n->{param}{cpan_bareword} 
+                  ? _mangle_ident( $n->{param}{cpan_bareword} )
+                  : _emit( $n->{param}{sub} );
+            my @a = split "-", $id;
+            my $version = ( @a > 1 && $a[-1] =~ /^[0-9]/ ? $a[-1] : '' );
+            return "extends '" . $a[0] . "'";
+        }
+        
+        if ( $n->{sub}{bareword} eq 'call' ) {
+            # call;
+            #warn "super call: ",Dumper $n;
+            return "super";  # param list?
+        }
+        
         if ( $n->{sub}{bareword} eq 'use' ) {
             # use v6-pugs
             if ( exists $n->{param}{cpan_bareword} ) {
@@ -203,65 +319,110 @@ sub default {
         return " print '', " . _emit( $n->{param} ) . ";\n" .
             " print " . '"\n"'
             if $n->{sub}{bareword} eq 'say';
+
+        # TODO - other builtins
+        return " (defined " . _emit( $n->{param} ) . ")"
+            if $n->{sub}{bareword} eq 'defined';
+            
+        # XXX: handle args
+        return "Pugs::Runtime::Perl6::Routine->new(Devel::Caller::caller_cv(1))"
+            if $n->{sub}{bareword} eq 'caller';
+
+        # ???
+        $n->{sub}{bareword} = 'die'
+            if $n->{sub}{bareword} eq 'fail';
+            
         return ' ' . _mangle_ident( $n->{sub}{bareword} ) . '(' . _emit( $n->{param} ) . ')';
     }
     
     if ( $n->{op1} eq 'method_call' ) {    
         #warn "method_call: ", Dumper( $n );
-        if ( $n->{method}{bareword} eq 'print' ||
-             $n->{method}{bareword} eq 'warn' ) {
+        if ( $n->{method}{dot_bareword} eq 'print' ||
+             $n->{method}{dot_bareword} eq 'warn' ) {
             my $s = _emit( $n->{self} );
             if ( $s eq _mangle_var('$*ERR') ) {  
                 return " print STDERR '', " . _emit( $n->{param} );
             }
             return " print '', $s";
         }
-        if ( $n->{method}{bareword} eq 'say' ) {
+        if ( $n->{method}{dot_bareword} eq 'say' ) {
             my $s = _emit( $n->{self} );
             if ( $s eq _mangle_var('$*ERR') ) { 
                 return " print STDERR '', " . _emit( $n->{param} ) . ', "\n"';
             }
             return " print '', $s" . ', "\n"';
         }
-        if ( $n->{method}{bareword} eq 'perl' ) {
+        if ( $n->{method}{dot_bareword} eq 'perl' ) {
             return 'Pugs::Runtime::Perl6::perl(' . _emit( $n->{self} ) . ")\n";
         }
-        #warn "method_call: ", Dumper( $n );
-        
-        # "autobox"
-        
-        if ( exists $n->{self}{code} ) {
-            # &code.goto;
-            return 
-                " \@_ = (" . _emit( $n->{param}, '  ' ) . ");\n" .
-                " " . _emit( $n->{method}, '  ' ) . " " .
-                    _emit( $n->{self}, '  ' );
+        # TODO: other builtins
+        if ( $n->{method}{dot_bareword} eq 'defined' ) {
+            return '(defined ' . _emit( $n->{self} ) . ")\n";
         }
         
-        if ( exists $n->{self}{scalar} ) {
-            # $scalar.++;
+        #warn "method_call: ", Dumper( $n );
+
+        # constructor
+        if ( exists $n->{self}{bareword} ) {
+            # Str.new;
             return 
-                " Pugs::Runtime::Perl6::Scalar::" . _emit( $n->{method}, '  ' ) . 
-                "(" . _emit( $n->{self}, '  ' ) .
-                ", " . _emit( $n->{param}, '  ' ) . ")" ;
+                " " . _emit( $n->{self} ) . "->" . _emit( $n->{method} ) . 
+                "(" . _emit( $n->{param} ) . ") ";
+        }
+    
+        # "autobox"
+        
+        if ( exists $n->{self}{code} && $n->{method}{dot_bareword} eq 'goto') {
+            # &code.goto;
+            return 
+                " \@_ = (" . _emit( $n->{param} ) . ");\n" .
+                " " . _emit( $n->{method} ) . " " .
+                    _emit( $n->{self} );
+        }
+        if ( exists $n->{self}{code} ) {
+            # &?ROUTINE.name;
+            return 
+                _emit( $n->{self} ) . "->" .
+                _emit( $n->{method} ) . "(" . _emit( $n->{param} ) . ")"
+        }
+        
+        #warn "method: ", Dumper( $n );
+        if ( exists $n->{self}{scalar} ) {
+            # $.scalar.method(@param)
+            return " " . _emit( $n->{self} ) . '->' .
+                _emit( $n->{method} ) .
+                '(' . _emit( $n->{param} ) . ')'
+                if $n->{self}{scalar} =~ /^\$\./;
+            
+            # $scalar.++;
+            # runtime decision - method or lib call
+            return 
+                "( Scalar::Util::blessed " . _emit( $n->{self} ) . " ? " .
+            
+                    _emit( $n->{self} ) . "->" . 
+                    _emit( $n->{method} ) . "(" . _emit( $n->{param} ) . ")" .
+                
+                " : " .
+                
+                    " Pugs::Runtime::Perl6::Scalar::" . _emit( $n->{method}, '  ' ) . 
+                    "(" . _emit( $n->{self} ) .
+                    ", " . _emit( $n->{param} ) . ")" .
+                
+                " )";
         }
         
         if ( exists $n->{self}{op1} ) {
             # %var<item>.++;
-            #warn "method: ", Dumper( $n );
-            return " " . _emit( $n->{method} ) .
-                '(' .
-                join ( ",\n", 
-                    map { _emit( $_ ) } ( $n->{self}, $n->{param} )
-                ) .
-                ')';
+            return
+                _emit( $n->{self} ) . "->" . 
+                _emit( $n->{method} ) . "(" . _emit( $n->{param} ) . ")";
         }
             
         # normal methods or subs
         
         return " " . _mangle_ident( $n->{sub}{bareword} ) .
             '(' .
-            join ( ";\n", 
+            join ( ";\n",   # XXX
                 map { _emit( $_ ) } @{$n->{param}} 
             ) .
             ')';
@@ -274,19 +435,21 @@ sub statement {
     my $n = $_[0];
     #warn "statement: ", Dumper( $n );
     
-    if ( $n->{op1}{stmt} eq 'if'     || 
-         $n->{op1}{stmt} eq 'unless' ) {
-        return  " " . $n->{op1}{stmt} . 
+    if ( $n->{statement} eq 'if'     || 
+         $n->{statement} eq 'unless' ) {
+        return  " " . $n->{statement} . 
                 '(' . _emit( $n->{exp1} ) . ')' .
                 " {\n" . _emit( $n->{exp2} ) . "\n }\n" .
                 " else" .
                 " {\n" . _emit( $n->{exp3} ) . "\n }";
     }
 
-    if ( $n->{op1}{stmt} eq 'sub' ) {
+    if ( $n->{statement} eq 'sub'       ||
+         $n->{statement} eq 'submethod' ||
+         $n->{statement} eq 'method'     ) {
         #warn "sub: ",Dumper $n;
 
-        my $name = _mangle_ident( $n->{name}{bareword} );
+        my $name = _mangle_ident( $n->{name} );
 
         my $export = '';
         for my $attr ( @{$n->{attribute}} ) {
@@ -297,43 +460,32 @@ sub statement {
         }
 
         return  $export .
-                " " . $n->{op1}{stmt} . 
-                ' ' . $name . 
-                " {\n" . 
+                " sub " . $name . 
+                " {\n" .
+                    (
+                        $n->{statement} =~ /method/
+                        ? " my \$self = shift; "   # default invocant 
+                        : ""
+                    ) .
                     _emit_parameter_binding( $n->{signature} ) .
                     _emit( $n->{block} ) . 
                 "\n }";
     }
 
-    if ( $n->{op1}{stmt} eq 'method' ) {
-        # Moose: sub clear { my $self = shift;';
-        # method clear {
-        #warn "method: ",Dumper $n;
-        return "sub " . _emit( $n->{name} ) .
-            " { my \$self = shift; " . 
-                _emit_parameter_binding( $n->{signature} ) .
-                _emit( $n->{block} ) . 
-            "\n }";
-    }
-
-
-    if ( $n->{op1}{stmt} eq 'for' ) {
-        #warn "sub: ",Dumper $n;
+    if ( $n->{statement} eq 'for' ) {
+        #warn "for: ",Dumper $n;
         if ( exists $n->{exp2}{pointy_block} ) {
-            return  " " . $n->{op1}{stmt} . 
-                    ' my ' . _emit( $n->{exp2}{signature} ) . '' . 
-                    ' (' . _emit( $n->{exp1} ) . ')' . 
-                    " {\n" . 
-                        # _emit_parameter_binding( $n->{signature} ) .
-                        _emit( $n->{exp2}{pointy_block} ) . 
-                    "\n }";
+            return  " " . $n->{statement} . 
+                    ( $n->{exp2}{signature} 
+                      ? ' my ' . _emit( $n->{exp2}{signature} ) 
+                      : '' 
+                    ) . 
+                    ' ( ' . _emit( $n->{exp1} ) . ' )' . 
+                    " { " . _emit( $n->{exp2}{pointy_block} ) . " }";
         }
-        return  " " . $n->{op1}{stmt} . 
-                ' (' . _emit( $n->{exp1} ) . ')' . 
-                " {\n" . 
-                    # _emit_parameter_binding( $n->{signature} ) .
-                    _emit( $n->{exp2} ) . 
-                "\n }";
+        return  " " . $n->{statement} . 
+                ' ( ' . _emit( $n->{exp1} ) . ' )' . 
+                " { " . _emit( $n->{exp2} ) . " }";
     }
 
     return _not_implemented( $n, "statement" );
@@ -341,17 +493,74 @@ sub statement {
 
 sub infix {
     my $n = $_[0];
-    # print "infix: ", Dumper( $n );
-    
+    #print "infix: ", Dumper( $n );
+
     if ( $n->{op1}{op} eq '~' ) {
         return _emit( $n->{exp1} ) . ' . ' . _emit( $n->{exp2} );
     }
+    if ( $n->{op1}{op} eq '~=' ) {
+        return _emit( $n->{exp1} ) . ' .= ' . _emit( $n->{exp2} );
+    }
+    if ( $n->{op1}{op} eq '//'  ||
+         $n->{op1}{op} eq 'err' ) {
+        return ' do { my $_tmp_ = ' . _emit( $n->{exp1} ) . 
+            '; defined $_tmp_ ? $_tmp_ : ' . _emit( $n->{exp2} ) . '}';
+    }
     
     if ( $n->{op1}{op} eq ':=' ) {
         #warn "bind: ", Dumper( $n );
-        return " tie " . _emit( $n->{exp1} ) . 
-            ", 'Pugs::Runtime::Perl6::Scalar::Alias', " .
-            "\\" . _emit( $n->{exp2} );
+        if ( exists $n->{exp2}{scalar} ) {
+            return " tie " . _emit( $n->{exp1} ) . 
+                ", 'Pugs::Runtime::Perl6::Scalar::Alias', " .
+                "\\" . _emit( $n->{exp2} );
+        }
+        else {
+            # XXX: for now, should use data::bind
+            return _emit( $n->{exp1}).' = '._emit( $n->{exp2});
+        }
+    }
+    if ( $n->{op1}{op} eq '~~' ) {
+        return _emit( $n->{exp1} ) . ' =~ (ref(' . _emit( $n->{exp2} ).') eq "REGEX" ? '._emit($n->{exp2}).' : quotemeta('._emit($n->{exp2}).'))';
+    }
+
+    if ( $n->{op1}{op} eq '=' ) {
+        # warn "{'='}: ", Dumper( $n );
+        if ( exists $n->{exp1}{scalar} ) {
+            #warn "set $n->{exp1}{scalar}";
+            return _var_set( $n->{exp1}{scalar} )->( _var_get( $n->{exp2} ) );
+        }
+        if ( exists $n->{exp1}{op1}  &&
+             $n->{exp1}{op1}{op} eq 'has' ) {
+            #warn "{'='}: ", Dumper( $n );
+            # XXX - changes the AST
+            push @{ $n->{exp1}{attribute} },
+                 [  { bareword => 'default' }, 
+                    $n->{exp2} 
+                 ]; 
+            #warn "{'='}: ", Dumper( $n );
+            return _emit( $n->{exp1} );
+        }
+        return _emit( $n->{exp1} ) . 
+            " = " . _var_get( $n->{exp2} );
+    }
+
+    if ( $n->{op1}{op} eq '+=' ) {
+        #warn "{'='}: ", Dumper( $n );
+        if ( exists $n->{exp1}{scalar} ) {
+            #warn "set $n->{exp1}{scalar}";
+            return _var_set( $n->{exp1}{scalar} )->( 
+                _emit(
+                  {
+                    fixity => 'infix',
+                    op1 => { op => '+' },
+                    exp1 => $n->{exp1},
+                    exp2 => $n->{exp2},
+                  }
+                )
+            );
+        }
+        return _emit( $n->{exp1} ) . 
+            " = " . _emit( $n->{exp2} );
     }
 
     if ( exists $n->{exp2}{bare_block} ) {
@@ -380,8 +589,18 @@ sub circumfix {
 
 sub postcircumfix {
     my $n = $_[0];
-    #print "postcircumfix: ", Dumper( $n );
+    #warn "postcircumfix: ", Dumper( $n );
     
+    if ( $n->{op1}{op} eq '(' &&
+         $n->{op2}{op} eq ')' ) {
+        # warn "postcircumfix:<( )> ", Dumper( $n );
+        # $.scalar(@param)
+        return " " . _emit( $n->{exp1} ) . 
+            '->(' . _emit( $n->{exp2} ) . ')'
+            if exists $n->{exp1}{scalar} &&
+               $n->{exp1}{scalar} =~ /^\$\./;
+    }
+            
     if ( $n->{op1}{op} eq '[' &&
          $n->{op2}{op} eq ']' ) {
 
@@ -410,10 +629,14 @@ sub prefix {
     my $n = $_[0];
     # print "prefix: ", Dumper( $n );
     
+    if ( $n->{op1}{op} eq ':' ) {
+        return _emit( $n->{exp1} ) . "  # XXX :\$var not implemented\n";
+    }
+    
     if ( $n->{op1}{op} eq 'my' ||
          $n->{op1}{op} eq 'our' ) {
-        die "not implemented 'attribute'",Dumper $n
-            if @{$n->{attribute}};
+        #die "not implemented 'attribute'",Dumper $n
+        #    if @{$n->{attribute}};
         return $n->{op1}{op} . ' ' . _emit( $n->{exp1} );
     }
 
@@ -421,15 +644,32 @@ sub prefix {
             # Moose: has 'xxx';
             # has $x;
             #warn "has: ",Dumper $n;
+            
+            my $name = _emit( $n->{exp1} );
+            #my $name = _emit( $n->{exp1} );
+            $name =~ s/^\$//;  # remove sigil
+            
+            my $raw_name;
+            $raw_name = $n->{exp1}{scalar} if exists $n->{exp1}{scalar};
+            $env{$raw_name}{set} = sub { 
+                "\$self->" . substr($raw_name,2) . "(" . $_[0] . ")" 
+            };
+            # is rw?
+            #warn Dumper @{$n->{attribute}};
+            my $is_rw = grep { $_->[0]{bareword} eq 'is' &&
+                               $_->[1]{bareword} eq 'rw' } @{$n->{attribute}};
+            $env{$raw_name}{set} = sub { 
+                "\$self->{'" . substr($raw_name,2) . "'} = " . $_[0] 
+            }
+                if $is_rw;
+            
             my $attr = join( ', ', 
                 map { 
                     join( ' => ', map { "'" . _emit($_) . "'" } @$_ )
                 } @{$n->{attribute}}
             );
 
-            my $name = _emit( $n->{exp1} );
-            $name =~ s/^\$//;  # remove sigil
-            return $n->{op1}{op} . " '$name' => ( $attr )";
+            return $n->{op1}{op} . " '" . substr($raw_name,2) . "' => ( $attr )";
     }
 
     if ( $n->{op1}{op} eq 'try' ) {
@@ -452,6 +692,12 @@ sub prefix {
             _mangle_var( '$!' ) . ' = $@; ' .
             '@result }';  # /do
     }
+    if ( $n->{op1}{op} eq '~' ) {
+        return ' "" . ' . _emit( $n->{exp1} );
+    }
+    if ( $n->{op1}{op} eq '!' ) {
+        return _emit( $n->{exp1} ) . ' ? 0 : 1 ';
+    }
     if ( $n->{op1}{op} eq '++' ||
          $n->{op1}{op} eq '--' ||
          $n->{op1}{op} eq '+'  ) {
@@ -479,4 +725,17 @@ sub postfix {
     return _not_implemented( $n, "postfix" );
 }
 
+sub ternary {
+    my $n = $_[0];
+    # print "ternary: ", Dumper( $n );
+
+    if ( $n->{op1}{op} eq '??' ||
+         $n->{op2}{op} eq '!!' ) {
+        return _emit( $n->{exp1} ) . 
+            ' ? ' . _emit( $n->{exp2} ) .
+            ' : ' . _emit( $n->{exp3} ) ;
+    }
+
+    return _not_implemented( $n, "ternary" );
+}
 1;
